/* Copyright (C) Duncan Macleod (2014-2021)
 *
 * This file is part of GWpy.
 *
 * GWpy is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * GWpy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GWpy.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * This file contains sphinx-specific extensions to the core GWpy CSS
 */

/* -- layout -------------------------------------------------------------- */

body {
		font-size: 15px;
		font-family: "Helvetica Neue",Helvetica,Arial,sans-serif;
}

.navbar + .container {
		padding-top: 20px;
}

/* -- navbar -------------------------------------------------------------- */

.navbar {
		background-color: #333;
}

/* -- footer -------------------------------------------------------------- */

html {
		position: relative;
		min-height: 100%;
}

body {
		margin-bottom: 140px;
}

.footer {
		position: absolute;
		bottom: 0;
		width: 100%;
		height: 100px;
		background-color: #eee;
		padding-top: 20px;
		padding-bottom: 20px;
}

/* -- side-bar formatting ------------------------------------------------- */

.bs-sidenav {
		background-color: transparent;
		border: 0;
		border-radius: 0;
		border-right: 1px solid #eee;
}

/* -- headers ------------------------------------------------------------- */

h1 {
		font-size: 32px;
}
h2 {
		font-size: 28px;
}
h3 {
		font-size: 24px;
}

/* -- links --------------------------------------------------------------- */

a.external,
a.external:hover,
a.external:focus {
		color: #337ab7;
}

.alert a.external,
.alert a.external:hover,
.alert a.external:focus {
		color: inherit;
		text-decoration: underline;
}

/* -- tables -------------------------------------------------------------- */

table > thead > tr > th,
.table > thead > tr > th,
table > tbody > tr > th,
.table > tbody > tr > th,
table > tfoot > tr > th,
.table > tfoot > tr > th,
table > thead > tr > td,
.table > thead > tr > td,
table > tbody > tr > td,
.table > tbody > tr > td,
table > tfoot > tr > td,
.table > tfoot > tr > td {
		border-top: 1px solid #ecf0f1;
}

table a {
		text-decoration: none !important;
}

/* -- code formatting ----------------------------------------------------- */

pre {
		background-color: #333;
}

code {
	padding: 0px
}

.alert-info code {
		color: #fccfcf;
}

/* -- autoclass formatting ------------------------------------------------ */

/* class title */
dl.class > dt > em,
dl.class > dt > code {
		font-size: 150%;
}
dl.class > dt > code.descclassname {
		padding-right: 0px;
}
dl.class > dt > code:descname {
		padding-left: 0px;
}

/* class documentation */
dl.method > dd,
dl.classmethod > dd,
dl.attribute > dd {
		margin-left: 40px;
}

/* parameter lists */
dl.class table.field-list,
dl.function table.field-list,
dl.method table.field-list {
		margin-bottom: 0px;
}

dl.class table.field-list th,
dl.function table.field-list th,
dl.method table.field-list th {
		padding-left: 0px;
}

dl.class table.field-list blockquote,
dl.function table.field-list blockquote,
dl.data table.field-list blockquote,
dl.method table.field-list blockquote {
		border-left: none;
		margin-bottom: 0px;
		padding: 0px 30px 10px;
		font-size: inherit;
}

/* table inside method */
dl.classmethod table {
		width: auto;
}

dl.classmethod table th:first-child,
dl.classmethod table td:first-child, {
		padding-left: 0px;
}

.class .rubric,
.classmethod .rubric,
.method .rubric {
		margin-top: 10px;
}

div.seealso dd {
		margin-left: 20px;
}
